#include "hash_test.h"

int verbose,timer,gnuplot,readmemory,counter = 0;
clock_t start,end;
void *buf;
void *currentbuf;
size_t filesize;

int main(int argc, char *argv[]){

	int fd;

	if(argv[1] == NULL){
		fprintf(stderr,"Error: no input file given?\n\n");
		usage(argv[0]);
		exit(1);
	}
	if(argv[2] == NULL){
		fprintf(stderr,"Error: no hash function is defined?\n");
		usage(argv[0]);
		exit(1);
	}
 
	if(argv[3] == NULL){
		fprintf(stderr, "Error: no verbose mode given?\n");
		usage(argv[0]);
		exit(1);
	}

	if(argv[4] == NULL){
		fprintf(stderr, "Error: no timer mode given?\n");
		usage(argv[0]);
		exit(1);
	}

	if(argv[5] == NULL){
		fprintf(stderr, "Error: no read into memory given?\n");
		usage(argv[0]);
		exit(1);
	}

	if(argv[6] == NULL){
		fprintf(stderr, "Error: Please give output format?\n");
		usage(argv[0]);
		exit(1);
	}

	// Open the input file for reading only
	fd = open(argv[1], O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "%s: unable to open file: %s\n", argv[0], argv[1]);
		exit(1);
	}

	// Find the file size
	stat(argv[1] ,&sbuf);  
	filesize = (size_t)sbuf.st_size;
  
	// Set verbose mode
	verbose = atoi(argv[3]);

	// Set timer mode
	timer = atoi(argv[4]);

	// Set read memory mod and read file into memory if given
	readmemory = atoi(argv[5]);
	if (readmemory) {
		printf("Starting to load file into memory....\n");
		buf = readintomemory(fd, filesize);
		currentbuf = buf;
		printf("Done loading file into memory....\n");
	} else {
		buf = NULL;
	}

	// Set output format 
	gnuplot = atoi(argv[6]);

  
	if (!verbose && !gnuplot) {
		printf("Starting hashing....\n");
	}

	// Chosse the current hash function
	if (strcmp(argv[2],"fnv16") == 0) {
		fnv16_test(fd); 
	} else if (strcmp(argv[2],"fnv16a") == 0) {
		fnv16a_test(fd);
	} else if (strcmp(argv[2],"fnv32") == 0) {
		fnv32_test(fd);
	} else if (strcmp(argv[2],"fnv32a") == 0) {
		fnv32a_test(fd);
	} else if (strcmp(argv[2],"fnv64") == 0) {
		fnv64_test(fd);
	} else if (strcmp(argv[2],"fnv64a") == 0) {
		fnv64a_test(fd);
	} else if (strcmp(argv[2],"bobs64a") == 0) {
		bobs64a_test(fd);
	} else if (strcmp(argv[2],"bobs64b") == 0) {
		bobs64b_test(fd);
	} else if (strcmp(argv[2],"bobs64c") == 0) {
		bobs64c_test(fd);
	} else if (strcmp(argv[2],"bobs32a") == 0) {
		bobs32a_test(fd);
	} else if (strcmp(argv[2],"bobs32b") == 0) {
		bobs32b_test(fd);
	} else if (strcmp(argv[2],"bobs32c") == 0) {
		bobs32c_test(fd);
	} else if (strcmp(argv[2],"SFH") == 0) {
		SFH_test(fd);
	} else if (strcmp(argv[2],"MD5") == 0) {
		md5_test(fd);
	} else if (strcmp(argv[2],"sha1") == 0) {
		sha1_test(fd);
	} else {
		printf("The selected hash function is not implemented: %s\n", argv[2]);
	}

	if (!verbose && !gnuplot) {
		printf("Hashing finished....\n");
	}

	// Close file and free memory
	close(fd);
	free(buf);

	return 0;
}

// Allocates memory and writes content of file in Returns : a void *
// pointing to the first element of memory containg the content of the
// file given
void *readintomemory( int fd, size_t filesize){
  
	void *buf;
	int readcnt = 0;
	buf = malloc(filesize);
  
	if (buf == NULL){
		printf("Alloc error in memory\n");
		exit(1);
	}
  
	readcnt = read(fd, buf, filesize);
	if (readcnt == 0){
		printf("Error reading file into memory\n");
		exit(1);
	} else {
		printf("Loaded %i bytes into memory\n", readcnt);
	}
  
	return buf;
}

// Findes out of the data is in memory or on disk and return
// information about one page
pageinfo_t readpage(int fd, void *adr) {
  
	int readcnt = 0;
	pageinfo_t pageinfo;
	unsigned char page[FRAMESIZE];

	if (readmemory) {
		if (currentbuf < (buf + filesize))
		{
			pageinfo.size = FRAMESIZE;
			pageinfo.data = currentbuf;
			currentbuf = currentbuf + FRAMESIZE;
		}
		else
		{
			pageinfo.size = 0;
		}
	} else {
		pageinfo.size = read(fd, page, FRAMESIZE);
		pageinfo.data = page;
	}

	return pageinfo;
}

// Print the usage information to the screeen
int usage (char *prg) {
	printf("Hash functions is: fnv16, fnv16a, fnv32, fnv32a, fnv64, fnv64a, bobs32a, bobs32b,\n");
	printf("                   bobs32c, bobs64a, bobs64b, bobs64c, MD5, sha1, SFH\n");
	printf("Usage: %s memdump hash_function 1|0 (verbose on/off) 1|0 (time on/off) 1|0 (Read into memory on/off) 1|0 (long output on/off)\n\n", prg);

	return 0;
}

// Time taking functions, this states a timer
void timer_start() {
	start = times(&tbuff);
}

// This function stoppes the timer and print the time spent
void timer_end() {
	end = times(&tbuff);

	printf("Real time = %4.2lf\n", ((double)(end-start))/CLK_TCK);
	printf("CPU used  = %4.2lf\n", ((double)tbuff.tms_utime)/CLK_TCK);
	printf("System    = %4.2lf\n", ((double)tbuff.tms_stime)/CLK_TCK);
}

/********************************************************************
 * Fnv hash function test implementations both in 32/32a and 64/64a *
 * bit versions                                                     *
 ********************************************************************/
int fnv16_test(int fd) {

	char hex_hval[10+1];
	Fnv32_t hval = FNV1_32_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	while(1) {
	  	  
		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = fnv_32_buf(page.data, page.size, FNV1_32_INIT);
		hval = (hval>>16) ^ (hval & MASK_16);
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%04lx", hval & (u_int32_t)0xffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int fnv16a_test(int fd) {

	char hex_hval[10+1];
	Fnv32_t hval = FNV1_32_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	while(1) {
	  	  
		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = fnv_32_buf(page.data, page.size, FNV1_32_INIT);
		hval = (hval>>16) ^ (hval & MASK_16);
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%04lx", hval & (u_int32_t)0xffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int fnv32_test(int fd) {

	char hex_hval[10+1];
	Fnv32_t hval = FNV1_32A_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	while(1) {
	  	  
		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = fnv_32a_buf(page.data, page.size, FNV1_32A_INIT);
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%08lx", hval & (Fnv32_t)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int fnv32a_test(int fd) {

	char hex_hval[10+1];
	Fnv32_t hval =  FNV1_32A_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = fnv_32a_buf(page.data, page.size, FNV1_32A_INIT);
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%08lx", hval & (Fnv32_t)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

// FNV 64/64a bit hash functions
int fnv64_test(int fd) {

	char hex_hval[17+1];
	Fnv64_t hval = FNV1_64_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = fnv_64_buf(page.data, page.size, FNV1_64_INIT);
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%016llx", hval & (Fnv64_t)0xffffffffffffffffULL);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i, %llu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int fnv64a_test(int fd) {

	char hex_hval[17+1];
	Fnv64_t hval = FNV1_64A_INIT;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}
	
	// Hash the data in the file
	while(1) {
		
		// Read in page
		page = readpage(fd, buf);
		
		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
		
		// Hash the page
		hval = fnv_64a_buf(page.data, page.size, FNV1_64A_INIT);
		
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%016llx", hval & (Fnv64_t)0xffffffffffffffffULL);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %llu\n", counter, hval);
			counter++;
		}
	}
  
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

/********************************************************************
 * Bob Jenkins hash function test implementations both in 32 and 64 *
 * bit versions                                                     *
 ********************************************************************/
int bobs64a_test(int fd) {

	unsigned char hex_hval[16+1];
	ub8 hval;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = hash(page.data, (ub8)page.size, (ub8)0);

		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x%x", hval & (ub8)0xffffffffffffffffULL);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %llu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int bobs64b_test(int fd) {
  
	unsigned char hex_hval[16+1];
	ub8 hval;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = hash2((ub8 *)page.data, (FRAMESIZE/sizeof(ub8)), (ub8)0);
    
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x%x", hval & (ub8)0xffffffffffffffffULL);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %llu\n", counter, hval);
			counter++;
		}
	}
  
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int bobs64c_test(int fd) {
  
	unsigned char hex_hval[16+1];
	ub8 hval;
	pageinfo_t page;
  
	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}
  
	// Hash the data in the file
	while(1) {
    
		// Read in page
		page = readpage(fd, buf);
    
		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		hval = hash3(page.data, (ub8)page.size, (ub8)0);
    
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x%x", hval & (ub8)0xffffffffffffffffULL);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %llu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int bobs32a_test(int fd) {

	unsigned char hex_hval[8+1];
	ub4 hval;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}
  
	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = hash4(page.data, (ub4)page.size, (ub4)0);

    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x", hval & (ub4)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
    
	}
  
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int bobs32b_test(int fd) {

	unsigned char hex_hval[8+1];
	ub4 hval;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = hash5((ub4 *)page.data, (FRAMESIZE/sizeof(ub4)), (ub4)0);

    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x", hval & (ub4)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}

	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

int bobs32c_test(int fd) {

	unsigned char hex_hval[8+1];
	ub4 hval;
	pageinfo_t page;
  
	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = hash6(page.data, (ub4)page.size, (ub4)0);
	
    
		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x", hval & (ub4)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}
  
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

/********************************************************************
 * SFH hash function test implementations                           *
 ********************************************************************/
int SFH_test(int fd) {

	unsigned char hex_hval[40+1];
	uint32_t hval;
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}

	// Hash the data in the file
	while(1) {

		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		hval = SuperFastHash(page.data, page.size);

		if (verbose) {
			// Conveter hval to hex
			sprintf(hex_hval, "0x%x%x", hval & (uint32_t)0xffffffff);
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
    
	}
  
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

/********************************************************************
 * MD5 hash function test implementations                           *
 ********************************************************************/
int md5_test(int fd) {

	int i = 0;
	unsigned char hex_hval[32+1];
	md5_state_t state;
	md5_byte_t hval[16];
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}
  
	// Hash the data in the file
	while(1) {
    
		// Read in page
		page = readpage(fd, buf);

		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}

		// Hash the page
		md5_init(&state);
		md5_append(&state, (const md5_byte_t *)page.data, page.size);
		md5_finish(&state, hval);

		if (verbose) {
			// Conveter hval to hex
			for (i = 0; i < 16; ++i) {
				sprintf(hex_hval + i * 2, "%02x", hval[i]);
			}
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}  
    
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}

/********************************************************************
 * SHA1 hash function test implementations                           *
 ********************************************************************/
int sha1_test(int fd) {

	int i = 0;
	unsigned char hex_hval[40+1];
	sha1_context ctx;
	unsigned char hval[20];
	pageinfo_t page;

	// Check if time is gonna be take
	if (timer) {
		timer_start();
	}
  
	// Hash the data in the file
	while(1) {
    
		// Read in page
		page = readpage(fd, buf);
    
		// Test if page was read in else end while loop
		if (!(page.size > 0)) {
			break;
		}
    
		// Hash the page
		sha1_starts(&ctx);
		sha1_update(&ctx, page.data, page.size);
		sha1_finish(&ctx, hval);
    
    
		if (verbose) {
			// Conveter hval to hex
			for (i = 0; i < 20; ++i) {
				sprintf(hex_hval + i * 2, "%02x", hval[i]);
			}
			printf("%s\n", hex_hval);
		} else if (gnuplot) {
			printf("%i %lu\n", counter, hval);
			counter++;
		}
	}
 
	// End timer if time taken
	if (timer) {
		timer_end();
	}
}
  
