#!/usr/bin/perl

use Digest::SHA1 qw(sha1 sha1_hex sha1_base64);

# Getting commandline parameters
my $number_of_input_files = shift;

for (1 .. $number_of_input_files) {
    push(@input_files, shift);
}

my $page_size = shift;
my $output = shift;

# Check that the command line parameters are given
if (($number_of_input_files == "") or ($#input_files+1 != $number_of_input_files) or ($page_size == "") or ($page_size < 0)) {
    print "Memory analyse of image files v2\n";
    print "Usage: memory_analyse.plx <number of input files> <the input files> <memory page size> <out>\n\n";
    print "Normal page size is 4096 bytes.\n";
    print "If \"out\" is set then output to file(s) <input_name>_output.dat\n\n";
    exit(1);
}

# Init varibales
my $page = "";
my $start_of_lastpage = 0;
my $digest = "";
my $share_pages = 0;
my $zore_page_digest = "vV/fa/WqfbEtjLak7gZq2tQdwNY";
my $output_file = "";
my %number_of_pages;
my %zero_page;
my %hash_statical_data;
my %hash_container;
my %hash_total;

# Run through the input files
foreach $input_file (@input_files) {

    # Open the current input file
    open(FILE, "<$input_file");
    binmode(FILE); 

    # Set the hash value to zero
    $zero_page{$input_file} = 0;

    # Goes through the file reading one page a the time
    while (read(FILE, $page, $page_size) != 0) {
	
	# Convert the page to a hex number for processing
	$hex = unpack("H*", $page);

	# Use hex value as key in hash
	$digest = sha1_base64($hex);
	${$hash_container{$input_file}{$digest}} =  ${$hash_container{$input_file}{$digest}} + 1;
	$hash_total{$digest} =  $hash_total{$digest} + 1;

	# Finde the number of zero pages
	if ("$digest" eq "$zore_page_digest") {
	    $zero_page{$input_file}++;
	}
	
	# Count the number of pages
	$number_of_pages{$input_file}++;
    }
    close(FILE);
}

# Run through the hash to check if more the one page hashed 
# to the same value
foreach $file (keys %hash_container) {
    
    # Open output file if specified
    if ($output eq "out") {
	$output_file = $file . "_output.dat";
	open(OUTPUT, "> $output_file");
    }
	
    foreach $data (keys %{$hash_container{$file}}) {
	$hash_statical_data{$file}{${$hash_container{$file}{$data}}}++;
	$share_pages += ${$hash_container{$file}{$data}}-1;

	# Print data to output file if specified
	if ($output eq "out") {
	    print OUTPUT ${$hash_container{$file}{$data}}-1 . "\n";
	}
    }
    
    # Print out statical data
    print "\n\n\tStatistical Data -- $file \n";
    print "------------------------------------------------------\n";
    print "Size of memory           : " . ($page_size/1024/1024) * $number_of_pages{$file} . " Mb \n";
    print "Number of page(s)        : " . $number_of_pages{$file} . "\n";
    print "Number of zero page(s)   : " . $zero_page{$file} . "\n";
    print "Number of shared page(s) : " . $share_pages . "\n";
    printf("Percent of share page(s) : %.1f \%\n",($share_pages/$number_of_pages{$file})*100);
    printf("Percent of zero page(s)  : %.1f \%\n\n",($zero_page{$file}/$number_of_pages{$file})*100);
    
#     # Statical information about page shares
#     my @keys_in_hash = keys %{$hash_statical_data{$file}};
#     print "There are x page(s) which y shares\n";
#     print "x \t = \t y\n";
#     foreach (sort {$a <=> $b} @keys_in_hash) {
# 	print "$hash_statical_data{$file}{$_} \t = \t $_\n";
#     }

    # Close output file
    if ($output eq "out") {
	print OUTPUT "ED\n$zero_page{$file}";
	close(OUTPUT);
    }
    
    # Reset variables
    $share_pages = 0;
}

# Finde the amount of shared pages between the input files
@files = keys(%hash_container);
if ($#files > 0) {

    my $num_of_page = 0;
    my $zpage = 0;
    my $spage = 0;

    # Open output file if specified
    if ($output eq "out") {
	$output_file =  "total_$files[0]_output.dat";
	open(OUTPUT, "> $output_file");
    }
    
    # Sum up the number for page(s) and zero page(s)
    foreach (@files) {
	$num_of_page += $number_of_pages{$_};
	$zpage += $zero_page{$_};
    }

    # Finde the number of shared pages overall
    foreach (keys %hash_total) {
	$spage += $hash_total{$_}-1;

	# Print data to output file if specified
	if ($output eq "out") {
	    print OUTPUT $hash_total{$_}-1 . "\n";
	}
    }

    # Print out statical data
    print "\n\n\tStatistical for all input files\n";
    print "------------------------------------------------------\n";
    print "Size of memory           : " . ($page_size/1024/1024) * $num_of_page . " Mb \n";
    print "Number of page(s)        : " . $num_of_page . "\n";
    print "Number of zero page(s)   : " . $zpage . "\n";
    print "Number of shared page(s) : " . $spage . "\n";
    printf("Percent of share page(s) : %.1f \%\n",($spage/$num_of_page)*100);
    printf("Percent of zero page(s)  : %.1f \%\n\n",($zpage/$num_of_page)*100);

#     # Write number of zero page and close output file if specified
#     if ($output eq "out") {
# 	print OUTPUT "ED\n$zpage";
# 	close(OUTPUT);
#     }
}
